# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by ./mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{BidiLRO}
# 
# Meaning: Bi-directional category 'LRO'
#
return <<'END';
202D		
END
