# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by ./mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCombiningHalfMarks} (and fuzzy permutations)
# 
# Meaning: Block 'Combining Half Marks'
#
return <<'END';
FE20	FE2F	Combining Half Marks
END
